import config from './config.js'
import cm from './lib/codemirror.js'
window.cm = cm

function Editor() {
  this.init = async () => {
    if(config.statusBarAtTop)
      document.body.appendChild(statusbar.element)

    console.log(cm, cm.EditorView)

    this.workspace = await Platform.getHome()
    statusbar.updateWorkspace()

    this.view = new cm.EditorView({
      state: await openLandingDocument()
        .catch(console.error),
      parent: document.body
    })

    if(!config.statusBarAtTop)
      document.body.appendChild(statusbar.element)

    this.doc.setView(this.view)
    this.view.focus()
  }

  // Opens a directory or a file
  this.open = path => {

  }

  this.openDocument = async (path) => {
    this.doc = await Doc.open(path)
      .catch(err => `Failed to open "${path}": ${err}`)
    this.view.setState(await this.doc.createState())
    this.doc.setView(this.view)
  }

  this.saveDocument = () => {
    let path = this.doc.getPath()

    Platform.writeText(path, this.view.state.doc.toString())
      .then(() => this.doc.boundPath = path)
      // TODO: User interface for errors like this
      .catch(err => 'Failed to save file:' + err)
  }

  this.setDocPath = path => {
    this.doc.setPath(path)
    statusbar.updateFilename()
  }

  // this.updateFilename = () => {
  //   this.doc.path = Platform.dirname(this.doc.path) + '/' + statusbar.filename.value
  // }

  const openLandingDocument = async () => {
    let doc = new Doc(config.landingDocument.render(), Platform.join(this.workspace, config.landingDocument.name))

    this.doc = doc
    statusbar.updateFilename()
    return await doc.createState()
      .catch(err => console.error("Could not open landing document", err))
  }

  this.doc = null
  this.workspace = '/'
}

function Statusbar() {
  this.selectFilename = () => {
    // lastFilename = this.filename.value

    let path = editor.doc.getPath()

    if(path.startsWith(editor.workspace))
      path = path.slice(editor.workspace.length + 1)

    this.filename.value = path
    this.filename.select()
    this.filename.selectionStart = path.length - Platform.filename(path).length
  }

  this.updateFilename = () => {
    this.filename.value = Platform.filename(editor.doc.getPath())
  }

  this.updateWorkspace = () => {
    this.workspace.value = Platform.filename(editor.workspace)
  }

  const onPromptKeydown = (event) => {
    switch(event.key) {
      case 'Enter':
      case 'Escape':
        event.preventDefault()
        editor.view.focus()
        break
    }
  }

  const onFilenameExit = () => {
    if(this.filename.value == '') {
      this.updateFilename()
    } else {
      let path = this.filename.value
      editor.setDocPath(Platform.absolute(path) ? path : Platform.join(editor.workspace, path))
    }
  }

  this.element = document.createElement('nav')
  this.filename = document.createElement('input')
  this.workspace = document.createElement('input')
  // let lastFilename = ''

  this.filename.id = "filename"
  this.filename.addEventListener('click', this.selectFilename)
  this.filename.addEventListener('keydown', onPromptKeydown)
  this.filename.addEventListener('focusout', onFilenameExit)

  this.workspace.id = "workspace"

  this.element.appendChild(this.filename)
  this.element.appendChild(this.workspace)
}

function Doc(content, initialPath) {
  // if(path) {
  //   let workspaces = Platform.storage.get('workspaces')
  // }

  this.createState = async () => {
    let extensions = [
      Doc.zen.of([
        cm.lineNumbers(),
        cm.highlightActiveLine(),
        cm.highlightActiveLineGutter(),
        cm.highlightSpecialChars(),
      ]),
      cm.foldGutter(),
      cm.drawSelection(),
      cm.dropCursor(),
      cm.EditorState.allowMultipleSelections.of(true),
      cm.indentOnInput(),
      cm.syntaxHighlighting(cm.defaultHighlightStyle, { fallback: true }),
      cm.bracketMatching(),
      cm.closeBrackets(),
      cm.autocompletion(),
      cm.rectangularSelection(),
      cm.crosshairCursor(),
      cm.highlightSelectionMatches(),
      Keymaps.default
    ]

    if(this.language)
      extensions.push(
        await langManager.resolveExtension(this.language)
          .then(extension => Doc.language.of(extension))
      )

    return cm.EditorState.create({
      extensions,
      doc: content
    })
  }

  this.setPath = newPath => {
    path = newPath
    this.setLanguage()
  }

  this.getPath = () =>
    path

  this.setView = (newView) => {
    view = newView
    this.setLanguage()
  }

  this.setLanguage = (newLanguage = getLanguage()) => {
    let oldLanguage = this.language
    this.language = newLanguage

    if(!view)
      return

    if(this.language == null) {
      setViewLanguageExtension([])
    } else if(oldLanguage == null || oldLanguage.name !== this.language.name) {
      langManager.resolveExtension(this.language)
        .then(setViewLanguageExtension)
        .catch(err => console.error('Failed to set language extension', err))
    }
  }

  const getLanguage = () => {
    let ext = Platform.ext(path)

    if(ext) {
      return langManager.exts[ext]
    }
  }

  const setViewLanguageExtension = (extension) => {
    view.dispatch({
      effects: Doc.language.reconfigure(extension)
    })
  }

  // This is the path in the statusbar
  let path = initialPath ?? ''
  let view = null
  // This is where doc state info is saved
  this.boundPath = initialPath
  this.language = getLanguage()
}

Doc.open = async (path) => {
  console.log('Opening ' + path)

  let content = await Platform.readText(path)
    .catch(console.error)

  console.log(content)

  return new Doc(content ?? '', path)
}

Doc.tabSize = new cm.Compartment
Doc.zen = new cm.Compartment
Doc.language = new cm.Compartment

const Keymaps = {
  default: cm.keymap.of([
    ...cm.closeBracketsKeymap,
    ...cm.defaultKeymap,
    ...cm.searchKeymap,
    ...cm.foldKeymap,
    ...cm.lintKeymap,

    // Alt completion keymap that merges with tab to indent
    { key: "Ctrl-Space", run: cm.startCompletion },
    { key: "Escape", run: cm.closeCompletion },
    { key: "ArrowDown", run: cm.moveCompletionSelection(true) },
    { key: "ArrowUp", run: cm.moveCompletionSelection(false) },
    { key: "PageDown", run: cm.moveCompletionSelection(true, "page") },
    { key: "PageUp", run: cm.moveCompletionSelection(false, "page") },
    { key: "Tab", run(view, event) { event.preventDefault(); cm.acceptCompletion(view) || cm.indentMore(view) }, shift: cm.indentLess },

    {
      key: config.binds.editPath,
      run({ state }, event) {
        event.preventDefault()
        let mainSelection = state.selection.main

        if(!mainSelection.empty)
          statusbar.filename.value = state.sliceDoc(mainSelection.from, mainSelection.to)
        statusbar.selectFilename()
      }
    },
    {
      key: config.binds.open,
      run({ state }) {
        let mainSelection = state.selection.main

        if(!mainSelection.empty) {
          let path = state.sliceDoc(mainSelection.from, mainSelection.to).trim()
          
          console.log(path)

          editor.setDocPath(
            Platform.absolute(path) ?
              path :
              Platform.join(Platform.dirname(editor.doc.getPath()), path)
          )

          statusbar.updateFilename()
        }

        editor.openDocument(editor.doc.getPath())
      }
    },
    {
      key: config.binds.save,
      run() {
        editor.saveDocument()
      }
    },

    {
      key: config.binds.reload,
      run() {
        window.location.href += ''
      }
    }
  ])
}

function LanguageManager() {
  this.resolveExtension = (language) =>
    import(language.import)
      .then(mod => language.createExtension(mod.default))

  this.exts = {}

  for(let language of config.languages) {
    for(let ext of language.exts)
      this.exts[ext] = language
  }
}

window.addEventListener('load', () => {
  Neutralino.init()
  window.langManager = new LanguageManager()
  window.editor = new Editor()
  window.statusbar = new Statusbar()
  editor.init()
})